﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.GeneralizedQuadraticAssignment.Common;
using HeuristicLab.Random;

namespace HeuristicLab.Problems.GeneralizedQuadraticAssignment {
  [Item("GQAPPathRelinking", "Operator that performs path relinking between two solutions. It is described in Mateus, G., Resende, M., and Silva, R. 2011. GRASP with path-relinking for the generalized quadratic assignment problem. Journal of Heuristics 17, Springer Netherlands, pp. 527-565.")]
  [StorableClass]
  public class GQAPPathRelinking : GQAPCrossover, IQualitiesAwareGQAPOperator, IDemandsAwareGQAPOperator, ICapacitiesAwareGQAPOperator,
  IWeightsAwareGQAPOperator, IDistancesAwareGQAPOperator, IInstallationCostsAwareGQAPOperator, ITransportationCostsAwareGQAPOperator,
  IOverbookedCapacityPenaltyAwareGQAPOperator {

    public ILookupParameter<BoolValue> MaximizationParameter {
      get { return (ILookupParameter<BoolValue>)Parameters["Maximization"]; }
    }
    public IScopeTreeLookupParameter<DoubleValue> QualityParameter {
      get { return (IScopeTreeLookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public IScopeTreeLookupParameter<DoubleValue> FlowDistanceQualityParameter {
      get { return (IScopeTreeLookupParameter<DoubleValue>)Parameters["FlowDistanceQuality"]; }
    }
    public IScopeTreeLookupParameter<DoubleValue> InstallationQualityParameter {
      get { return (IScopeTreeLookupParameter<DoubleValue>)Parameters["InstallationQuality"]; }
    }
    public IScopeTreeLookupParameter<DoubleValue> OverbookedCapacityParameter {
      get { return (IScopeTreeLookupParameter<DoubleValue>)Parameters["OverbookedCapacity"]; }
    }
    public ILookupParameter<DoubleArray> DemandsParameter {
      get { return (ILookupParameter<DoubleArray>)Parameters["Demands"]; }
    }
    public ILookupParameter<DoubleArray> CapacitiesParameter {
      get { return (ILookupParameter<DoubleArray>)Parameters["Capacities"]; }
    }
    public ILookupParameter<DoubleMatrix> WeightsParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["Weights"]; }
    }
    public ILookupParameter<DoubleMatrix> DistancesParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["Distances"]; }
    }
    public ILookupParameter<DoubleMatrix> InstallationCostsParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["InstallationCosts"]; }
    }
    public IValueLookupParameter<DoubleValue> TransportationCostsParameter {
      get { return (IValueLookupParameter<DoubleValue>)Parameters["TransportationCosts"]; }
    }
    public IValueLookupParameter<DoubleValue> OverbookedCapacityPenaltyParameter {
      get { return (IValueLookupParameter<DoubleValue>)Parameters["OverbookedCapacityPenalty"]; }
    }

    public IValueParameter<PercentValue> CandidateSizeFactorParameter {
      get { return (IValueParameter<PercentValue>)Parameters["CandidateSizeFactor"]; }
    }

    [StorableConstructor]
    protected GQAPPathRelinking(bool deserializing) : base(deserializing) { }
    protected GQAPPathRelinking(GQAPPathRelinking original, Cloner cloner) : base(original, cloner) { }
    public GQAPPathRelinking()
      : base() {
      Parameters.Add(new LookupParameter<BoolValue>("Maximization", GeneralizedQuadraticAssignmentProblem.MaximizationDescription));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Quality", GQAPEvaluator.QualityDescription));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("FlowDistanceQuality", GQAPEvaluator.FlowDistanceQualityDescription));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("InstallationQuality", GQAPEvaluator.InstallationQualityDescription));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("OverbookedCapacity", GQAPEvaluator.OverbookedCapacityDescription));
      Parameters.Add(new LookupParameter<DoubleArray>("Demands", GeneralizedQuadraticAssignmentProblem.DemandsDescription));
      Parameters.Add(new LookupParameter<DoubleArray>("Capacities", GeneralizedQuadraticAssignmentProblem.CapacitiesDescription));
      Parameters.Add(new LookupParameter<DoubleMatrix>("Weights", GeneralizedQuadraticAssignmentProblem.WeightsDescription));
      Parameters.Add(new LookupParameter<DoubleMatrix>("Distances", GeneralizedQuadraticAssignmentProblem.DistancesDescription));
      Parameters.Add(new LookupParameter<DoubleMatrix>("InstallationCosts", GeneralizedQuadraticAssignmentProblem.InstallationCostsDescription));
      Parameters.Add(new ValueLookupParameter<DoubleValue>("TransportationCosts", GeneralizedQuadraticAssignmentProblem.TransportationCostsDescription));
      Parameters.Add(new ValueLookupParameter<DoubleValue>("OverbookedCapacityPenalty", GeneralizedQuadraticAssignmentProblem.OverbookedCapacityPenaltyDescription));
      Parameters.Add(new ValueParameter<PercentValue>("CandidateSizeFactor", "(η) Determines the size of the set of feasible moves in each path-relinking step relative to the maximum size. A value of 50% means that only half of all possible moves are considered each step.", new PercentValue(0.5)));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new GQAPPathRelinking(this, cloner);
    }

    public static IntegerVector Apply(IRandom random, ItemArray<IntegerVector> parents, BoolValue maximization, ItemArray<DoubleValue> qualities,
      ItemArray<DoubleValue> flowDistanceQualities, ItemArray<DoubleValue> installationQualities, ItemArray<DoubleValue> overbookedCapacities,
      DoubleArray demands, DoubleArray capacities, DoubleMatrix weights, DoubleMatrix distances, DoubleMatrix installationCosts,
      DoubleValue transportationCosts, DoubleValue overbookedCapacityPenalty, DoubleValue candidateSizeFactor) {
      if (random == null) throw new ArgumentNullException("random", "No IRandom provider is given.");
      if (parents == null || !parents.Any()) throw new ArgumentException("No parents given for path relinking.", "parents");
      if (parents.Length != 2) throw new ArgumentException(String.Format("Two parents were expected for path relinking, but {0} was/were given.", parents.Length.ToString()), "parents");
      if (parents[0].Length != parents[1].Length) throw new ArgumentException("The length of the parents is not equal.", "parents");
      if (qualities == null || qualities.Length == 0) throw new ArgumentException("The qualities are not given.", "qualities");
      if (qualities.Length != parents.Length) throw new ArgumentException(String.Format("There are a different number of parents ({0}) than quality values ({1})", parents.Length.ToString(), qualities.Length.ToString()));

      var source = parents[0];
      var target = parents[1];
      IntegerVector result;
      double resultQuality, resultFDQ, resultIQ, resultOC;

      int betterParent = 1;
      if ((maximization.Value && qualities[0].Value >= qualities[1].Value)
        || (!maximization.Value && qualities[0].Value <= qualities[1].Value))
        betterParent = 0;

      result = (IntegerVector)parents[betterParent].Clone();
      resultQuality = qualities[betterParent].Value;
      resultFDQ = flowDistanceQualities[betterParent].Value;
      resultIQ = installationQualities[betterParent].Value;
      resultOC = overbookedCapacities[betterParent].Value;

      var pi_prime = (IntegerVector)source.Clone();
      var fix = new HashSet<int>();
      var nonFix = new HashSet<int>(Enumerable.Range(0, demands.Length));
      var phi = new HashSet<int>(IntegerVectorEqualityComparer.GetDifferingIndices(pi_prime, target));

      while (phi.Any()) {
        var B = new List<GQAPSolution>();
        foreach (var v in phi) {
          int oldLocation = pi_prime[v];
          pi_prime[v] = target[v];
          var pi2 = makeFeasible(random, pi_prime, v, nonFix, demands, capacities);
          pi_prime[v] = oldLocation;

          double currentFDQ, currentIQ, currentOC;
          GQAPEvaluator.Evaluate(pi2, weights, distances, installationCosts, demands, capacities,
            out currentFDQ, out currentIQ, out currentOC);

          if (currentOC <= 0.0) {
            var quality = GQAPEvaluator.GetCombinedQuality(currentFDQ, currentIQ, currentOC,
                transportationCosts.Value, overbookedCapacityPenalty.Value);
            var solution = new GQAPSolution(pi2, new DoubleValue(quality), new DoubleValue(currentFDQ),
              new DoubleValue(currentIQ), new DoubleValue(currentOC));
            if (B.Count >= candidateSizeFactor.Value * phi.Count) {
              int mostSimilarIndex = -1;
              double mostSimilarValue = 1.0;
              for (int i = 0; i < B.Count; i++) {
                if (IsBetter(maximization.Value, solution.Quality.Value, B[i].Quality.Value)) {
                  var similarity = IntegerVectorEqualityComparer.GetSimilarity(solution.Assignment, B[i].Assignment);
                  if (similarity == 1.0) {
                    mostSimilarIndex = -1;
                    break;
                  } else if (similarity < mostSimilarValue) {
                    mostSimilarValue = similarity;
                    mostSimilarIndex = i;
                  }
                }
              }
              if (mostSimilarIndex >= 0)
                B[mostSimilarIndex] = solution;
            } else {
              bool contains = false;
              foreach (var b in B) {
                if (!IntegerVectorEqualityComparer.GetDifferingIndices(solution.Assignment, b.Assignment).Any()) {
                  contains = true;
                  break;
                }
              }
              if (!contains) B.Add(solution);
            }
          }
        }
        if (B.Any()) {
          GQAPSolution pi;
          if (maximization.Value) pi = B.SampleProportional(random, 1, B.Select(x => x.Quality.Value), false).First();
          else pi = B.SampleProportional(random, 1, B.Select(x => 1.0 / x.Quality.Value), false).First();
          var diff = IntegerVectorEqualityComparer.GetDifferingIndices(pi.Assignment, target);
          var I = phi.Except(diff);
          var i = I.SampleRandom(random);
          fix.Add(i);
          nonFix.Remove(i);
          pi_prime = pi.Assignment;
          if (IsBetter(maximization.Value, pi.Quality.Value, resultQuality)) {
            result = pi.Assignment;
            resultQuality = pi.Quality.Value;
            resultFDQ = pi.FlowDistanceQuality.Value;
            resultIQ = pi.InstallationQuality.Value;
            resultOC = pi.OverbookedCapacity.Value;
          }
        } else return result;
        phi = new HashSet<int>(IntegerVectorEqualityComparer.GetDifferingIndices(pi_prime, target));
      }

      return result;
    }

    protected override IntegerVector Cross(IRandom random, ItemArray<IntegerVector> parents) {
      return Apply(random, parents, MaximizationParameter.ActualValue, QualityParameter.ActualValue, FlowDistanceQualityParameter.ActualValue,
        InstallationQualityParameter.ActualValue, OverbookedCapacityParameter.ActualValue, DemandsParameter.ActualValue,
        CapacitiesParameter.ActualValue, WeightsParameter.ActualValue, DistancesParameter.ActualValue,
        InstallationCostsParameter.ActualValue, TransportationCostsParameter.ActualValue,
        OverbookedCapacityPenaltyParameter.ActualValue, CandidateSizeFactorParameter.Value);
    }

    private static IntegerVector makeFeasible(IRandom random, IntegerVector assignment, int equipment, HashSet<int> nonFix, DoubleArray demands, DoubleArray capacities, int maximumTries = 1000) {
      int l = assignment[equipment];
      var slack = ComputeSlack(assignment, demands, capacities);
      if (slack[l] >= 0) return (IntegerVector)assignment.Clone();

      IntegerVector result = null;
      int k = 0;
      while (k < maximumTries && slack[l] < 0) {
        result = (IntegerVector)assignment.Clone();
        do {
          var maxSlack = slack.Max();
          var T = nonFix.Where(x => x != equipment && assignment[x] == l && demands[x] <= maxSlack);
          if (T.Any()) {
            int i = T.SampleProportional(random, 1, T.Select(x => demands[x]), false).First();
            var j = Enumerable.Range(0, capacities.Length).Where(x => slack[x] >= demands[i]).SampleRandom(random);
            result[i] = j;
            slack[j] -= demands[i];
            slack[l] += demands[i];
          } else break;
        } while (slack[l] < 0);
        k++;
      }
      return result;
    }

    private static DoubleArray ComputeSlack(IntegerVector assignment, DoubleArray demands, DoubleArray capacities) {
      var slack = (DoubleArray)capacities.Clone();
      for (int i = 0; i < assignment.Length; i++) {
        slack[assignment[i]] -= demands[i];
      }
      return slack;
    }

    private static bool IsBetter(bool maximization, double quality, double otherQuality) {
      return maximization && quality > otherQuality || !maximization && quality < otherQuality;
    }
  }
}
